# Copyright (c) 2011 Valve Corporation 
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# 
# ETW (Event Tracing for Windows) profiling helpers.
# This allows easy insertion of Generic Event markers into ETW/xperf tracing
# which then aids in analyzing the traces and finding performance problems.

import os
import re
import shutil
import sys

if len(sys.argv) < 3:
	print "Syntax: PackETWSymbols ETWFilename.etl destdirname [-verbose]"
	print "This script looks for symbols needed to decode the specified trace, and"
	print "copies them to the specified directory. This allows moving traces to"
	print "other machines for analysis and sharing."
	sys.exit(0)

ETLName = sys.argv[1]
DestDirName = sys.argv[2]
if not os.path.exists(DestDirName):
	os.mkdir(DestDirName)

verbose = False
if len(sys.argv) > 3 and sys.argv[3].lower() == "-verbose":
	verbose = True

print "Extracting symbols from ETL file '%s'." % ETLName

# This command is slow but thorough -- it tries to build the symbol cache.
#command = "xperf.exe -i \"%s\" -tle -symbols -a symcache -quiet -build -imageid -dbgid" % ETLName
# This command is faster. It relies on symbols being loaded already for the modules of interest.
command = "xperf.exe -i \"%s\" -tle -a symcache -quiet -imageid -dbgid" % ETLName

print "Executing command '%s'" % command
lines = os.popen(command).readlines()

if len(lines) < 30:
	print "Error:"
	for line in lines:
		print line,
	sys.exit(0)

# Typical output lines (including one heading) look like this:
#TimeDateStamp,  ImageSize, OrigFileName, CodeView Record
#   0x4da89d03, 0x00bcb000, "client.dll", "[RSDS] PdbSig: {7b2a9028-87cd-448d-8500-1a18cdcf6166}; Age: 753; Pdb: u:\buildbot\dota_staging_win32\build\src\game\client\Release_dota\client.pdb"

scan = re.compile(r'   0x(.*), 0x(.*), "(.*)", "\[RSDS\].*; Pdb: (.*)"')

matchCount = 0
matchExists = 0
ourModuleCount = 0

# Get the users build directory
vgame = os.getenv("vgame")
vgame = vgame[:-5].lower()

prefixes = ["u:\\", "e:\\build_slave", vgame]

print "Looking for symbols built to:"
for prefix in prefixes:
	print "    %s" % prefix

# Default to looking for the SymCache on the C drive
prefix = "c"
# Look for a drive letter in the ETL Name and use that if present
if len(ETLName) > 1 and ETLName[1] == ':':
	prefix = ETLName[0]
else:
	# If there's no drive letter in the ETL name then look for one
	# in the current working directory.
	curwd = os.getcwd()
	if len(curwd) > 1 and curwd[1] == ':':
		prefix = curwd[0]

symCachePathBase = os.getenv("_NT_SYMCACHE_PATH");
if symCachePathBase == None or len(symCachePathBase) == 0:
	symCachePathBase = "%s:\\symcache\\" % prefix
elif symCachePathBase[-1] != '\\':
	symCachePathBase += '\\'

for line in lines:
	result = scan.match(line)
	if result is not None:
		#print result.groups()
		matchCount += 1
		TimeDateStamp = result.groups()[0]
		ImageSize = result.groups()[1]
		OrigFileName = result.groups()[2]
		PDBPath = result.groups()[3].lower()

		# Find out which PDBs are 'interesting'. There is no obvious heuristic
		# for this, but having a list of prefixes seems like a good start.
		ours = False
		for prefix in prefixes:
			if PDBPath.startswith(prefix):
				ours = True
		if ours:
			ourModuleCount += 1
			ours = True
			symFilePath = OrigFileName + "-" + TimeDateStamp + ImageSize + "v1.symcache"
			symCachePath = symCachePathBase + symFilePath
			if os.path.isfile(symCachePath):
				matchExists += 1
				print "Copying %s" % symCachePath
				shutil.copyfile(symCachePath, DestDirName + "\\" + symFilePath)
			else:
				print "Symbols for '%s' are not in %s" % (OrigFileName, symCachePathBase)
		else:
			#This is normally too verbose
			if verbose:
				print "Skipping %s" % PDBPath

print "%d symbol files found in the trace, %d appear to be ours, and %d of those exist in symcache." % (matchCount, ourModuleCount, matchExists)
