// Copyright (c) 2011 Valve Corporation 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal in
// the Software without restriction, including without limitation the rights to use,
// copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
// Software, and to permit persons to whom the Software is furnished to do so,
// subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
// IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// This exercises the functions in etwprof.cpp
//
// For more information see http://randomascii.wordpress.com/category/xperf/

#include "stdafx.h"
#include "etwprof.h"

// Pause for a short period of time with the CPU idle.
__declspec(noinline) void IdleDelay()
{
	ETWMark("Going to sleep for a bit...");
	Sleep(10);
}

// Pause for a short period of time with the CPU busy.
__declspec(noinline) void BusyDelay()
{
	ETWMark("Wasting CPU time for a bit...");
	DWORD startTick = GetTickCount();

	for (;;)
	{
		DWORD elapsed = GetTickCount() - startTick;
		if (elapsed > 10)
			break;
	}
}

__declspec(noinline) void TestMainProvider(int argc)
{
	IdleDelay();
	ETWMark("After an idle delay");

	IdleDelay();
	ETWMarkPrintf("argc is %d", argc);

	// Record Begin/End event pairs. Normally this is best
	// done with CETWScope.
	int64 startTick = ETWBegin("Busy work...");
	BusyDelay();
	ETWEnd("Busy work...", startTick);
}

__declspec(noinline) void TestWorkerProvider(int argc)
{
	IdleDelay();
	ETWWorkerMark("After an idle delay");

	IdleDelay();
	ETWWorkerMarkPrintf("argc is %d", argc);

	// Record Begin/End event pairs. Normally this is best
	// done with CETWScope.
	int64 startTick = ETWWorkerBegin("Busy work...");
	BusyDelay();
	ETWWorkerEnd("Busy work...", startTick);
}

__declspec(noinline) void TestHeapProvider()
{
	CETWScope timer("Lots of memory allocations");
	void* allocations[2400];

	const size_t delayInterval = 32;
	for (size_t i = 0; i < _countof(allocations); ++i)
	{
		// Allocate a random amount of memory.
		allocations[i] = malloc(1024 + rand());
		// Put in occasional delays to make for a better looking graph over time
		if ((i % delayInterval) == 0)
			IdleDelay();
	}

	// Free some of the memory, but not all. This allows demonstration
	// of AIFI (Allocated Inside Freed Inside) versus AIFO (Allocated Inside
	// Freed Outside) types in summary tables.
	for (size_t i = 0; i < _countof(allocations); ++i)
	{
		// 50% chance of freeing each block
		if (rand() < RAND_MAX / 2)
		{
			free(allocations[i]);
		}
		// Put in occasional delays to make for a better looking graph over time
		if ((i % delayInterval) == 0)
			IdleDelay();
	}


	// Put in a long enough delay so that the graph will show the allocaiton count
	// stabilizing before zeroing out at process exit.
	for (int i = 0; i < 20; ++i)
		IdleDelay();
}

int _tmain(int argc, _TCHAR* argv[])
{
	printf("Emitting ETW events...\n");

	// Insert named Begin/End events to mark the start and stop
	// of main().
	CETWScope globalTimer(__FUNCTION__);

	TestMainProvider(argc);
	TestWorkerProvider(argc);

	// Test the render provider.
	for (int i = 0; i < 40; ++i)
	{
		ETWRenderFrameMark();
		IdleDelay();
	}

	// Test the input provider
	// The first parameter indicates which button -- conventionally this is
	// zero for left, 1 for middle, and 2 for right.
	ETWMouseDown(2, 0, 10, 20);
	ETWMouseUp(2, 0, 10, 20);

	TestHeapProvider();

	printf("Done.\n");
	return 0;
}
